/* See LICENSE for licensing and NOTICE for copyright. */
package edu.vt.middleware.ldap.ed.jetty.realm;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import javax.security.auth.Subject;
import javax.servlet.ServletRequest;
import edu.vt.middleware.ldap.ed.EdAuth;
import edu.vt.middleware.ldap.ed.EdContext;
import org.eclipse.jetty.security.DefaultIdentityService;
import org.eclipse.jetty.security.IdentityService;
import org.eclipse.jetty.security.LoginService;
import org.eclipse.jetty.security.MappedLoginService.RolePrincipal;
import org.eclipse.jetty.security.MappedLoginService.UserPrincipal;
import org.eclipse.jetty.server.UserIdentity;
import org.eclipse.jetty.util.component.AbstractLifeCycle;
import org.ldaptive.Credential;
import org.ldaptive.DnParser;
import org.ldaptive.LdapEntry;
import org.ldaptive.LdapException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@link EdAuthRealm} to provide drop in authentication against EdAuth using Jetty security realms.
 *
 * @author  Middleware Services
 */
public class EdAuthRealm extends AbstractLifeCycle implements LoginService
{

  /** Class logger instance. */
  protected final Logger logger = LoggerFactory.getLogger(EdAuthRealm.class);

  /** Default IdentityService for realm. */
  private final IdentityService identityService = new DefaultIdentityService();

  /** Descriptive information about this Realm implementation. */
  private String name = "EdAuthRealm";

  /**
   * Returns the name of the realm. (default: EdAuthRealm)
   *
   * @return  {@link String}
   */
  @Override
  public String getName()
  {
    return name;
  }

  /**
   * Sets the name of the realm.
   *
   * @param  realmName  {@link String}
   */
  public void setName(final String realmName)
  {
    name = realmName;
  }

  /**
   * The core function which authenticates the user and gathers the roles.
   *
   * @param  username  {@link String}
   * @param  credentials  {@link Object}
   * @param  servletRequest {@link ServletRequest}
   *
   * @return  {@link UserIdentity} if the user is authenticated. returns <b>null</b> otherwise.
   */
  @Override
  public UserIdentity login(final String username, final Object credentials, final ServletRequest servletRequest)
  {
    UserIdentity user = null;
    final EdAuth auth = new EdAuth();
    try {
      final LdapEntry userEntry =
              auth.authenticate(username, new Credential(credentials.toString()), EdContext.AFFILIATION_ATTR,
                      EdContext.GROUP_MEMBERSHIP_ATTR);
      // add EdAuth specific affiliations to roles
      final List<String> userRoles = new ArrayList<>();
      final String[] affil = auth.getAffiliations(userEntry);
      if (affil != null) {
        Collections.addAll(userRoles, affil);
      }
      // add EdAuth specific group membership to roles
      final String[] groupMembership = auth.getGroupMembership(userEntry);
      if (groupMembership != null) {
        Arrays.stream(groupMembership).forEach(
          groupMember -> userRoles.add(DnParser.getValue(groupMember, EdContext.UUGID_ATTR)));
      }
      final UserPrincipal authenticatedPrincipal = new UserPrincipal() {
        /**
         * serialVersionUID
         */
        private static final long serialVersionUID = 2954316850743275483L;
        @Override
        public boolean authenticate(final Object credentials)
        {
          return true;
        }

        @Override
        public boolean isAuthenticated()
        {
          return true;
        }

        @Override
        public String getName()
        {
          return username;
        }
      };
      final Subject subject = new Subject();
      subject.getPrincipals().add(authenticatedPrincipal);
      subject.getPrivateCredentials().add(null);

      userRoles.stream().forEach(role -> subject.getPrincipals().add(new RolePrincipal(role)));
      subject.setReadOnly();
      user = identityService.newUserIdentity(
        subject,
        authenticatedPrincipal,
        userRoles.toArray(new String[userRoles.size()]));
    } catch (LdapException e) {
      logger.error("Could not authenticate user in realm.", e);
    }
    return user;
  }

  /**
   * This method will return true is the user is not null, false otherwise.
   *
   * @param  user  {@link UserIdentity}
   *
   * @return  {@link boolean}
   */
  @Override
  public boolean validate(final UserIdentity user)
  {
    return user != null;
  }

  /**
   * This method is not supported by this Realm. It will do nothing.
   *
   * @param  user  {@link UserIdentity}
   */
  @Override
  public void logout(final UserIdentity user)
  {
    // Do nothing.
  }

  /**
   * Return the default {@link IdentityService}
   *
   * @return  {@link IdentityService}
   */
  @Override
  public IdentityService getIdentityService()
  {
    return identityService;
  }

  /**
   * This method is not supported in this realm.
   *
   * @param  iService  {@link IdentityService}
   */
  @Override
  public void setIdentityService(final IdentityService iService)
  {
    // Do nothing.
  }

}
